unit UNews;

{$mode delphi}{$H+}

interface

uses 
  USqlDatabase, SysUtils, UCore, Classes, UDatabase;

type

  { TNewsCategory }

  TNewsCategory = class
    Id: Integer;
    Title: string;
    Permission: Integer;
    procedure LoadFromDbRecord(DbRow: TAssocArray);
    procedure Store;
  end;
  
  { TNewsItem }

  TNewsItem = class
    Id: Integer;
    Title: string;
    Content: string;
    Time: TDateTime;
    Author: string;
    IPAddress: string;
    Category: Integer;
    procedure LoadFromDbRecord(DbRow: TAssocArray);
    procedure Store;
  end;

  { TNews }

  TNews = class
    NewsCategoryList: TList;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromDb;
    function ShowNews(Category: Integer): string;
  end;

  function GenerateRSS: string;

implementation

uses
  URSS, pwmain, dateutils;

{ TNewsItem }

procedure TNewsItem.LoadFromDbRecord(DbRow: TAssocArray);
begin
  Id := StrToInt(DbRow.Values['id']);
  Title := DbRow.Values['title'];
  Content := DbRow.Values['content'];
  Time := UnixToDateTime(StrToInt(DbRow.Values['time']));
  Author := DbRow.Values['author'];
  IPAddress := DbRow.Values['ip'];
  Category := StrToInt(DbRow.Values['category']);
end;

procedure TNewsItem.Store;
var
  Values: TAssocArray;
begin
  Values := TAssocArray.Create;
  Values.AddKeyValue('content', Content);
  Values.AddKeyValue('title', Title);
  Values.AddKeyValue('author', Author);
  Values.AddKeyValue('ip', IPAddress);
  Values.AddKeyValue('time', IntToStr(DateTimeToUnix(Time)));
  Values.AddKeyValue('category', IntToStr(Category));
  Database.Update('news', Values, 'id=' + IntToStr(Id));
  Values.Free;
  inherited;
end;

function GenerateRSS: string;
const
  MaxNewsItemCount = 15;
//  UploadedFilesFolder = 'uploads/';
var
  RSSChannel:  TRSSChannel;
  Select: string;
  SelectParts: TArrayOfString;
  Where: string;
  DbRows: TDbRows;
  NewItem: TRSSChannelItem;
  I: Integer;
  News: TNews;
  NewsItem: TNewsItem;
begin
  News := TNews.Create;
  News.LoadFromDb;
  NewsItem := TNewsItem.Create;
  RSSChannel := TRSSChannel.Create;
  with RSSChannel do begin
    Title := 'ZděchovNET - Aktuality';
    Link := 'http://centrala.zdechov.net/';
    Description := 'Aktuality komunitní počítačové sítě ZděchovNET';
    WebMasterEmail := 'robie@centrum.cz';

(*
$Category = '';
$CategoryOption = '';
$CategoryOptionURL = '';
$CategoryName = '';
*)

  if GetCgiVar('select') = '' then Select := '' else Select := GetCgiVar('select');
  SelectParts := Explode('-', Select);
  Where := '';
  for I := 0 to High(SelectParts) do
    Where := Where + 'OR (category=' + SelectParts[I] + ')';
  Where := Copy(Where, 3, Length(Where));
  if Where = '' then Where := '1';

  DbRows := Database.Query('SELECT *,UNIX_TIMESTAMP(date) as time FROM news WHERE ' + Where + ' ORDER BY date DESC LIMIT 0,' + IntToStr(MaxNewsItemCount));  for I := 0 to DbRows.Count - 1 do begin
    NewsItem.LoadFromDbRecord(DbRows[I]);
    NewItem := TRSSChannelItem.Create;
    with NewItem do begin
(*
    EnclosuresText := '';
    if Values['enclosure'] <> '' then begin
      EnclosuresText := EnclosureText + '<br />Přílohy: ';
      Enclosures := explode(';', $Row['enclosure']);
    foreach($Enclosures as $Enclosure)
    {
      if(file_exists($UploadedFilesFolder.$Enclosure)) $EnclosuresText .= ' <a href="http://centrala.zdechov.net/aktuality/'.$UploadedFilesFolder.$Enclosure.'">'.$Enclosure.'</a>';
    }
  }
*)
      Title := TNewsCategory(News.NewsCategoryList[NewsItem.Category]).Title + ' - ' + NewsItem.Title;
      Link := 'http://centrala.zdechov.net/aktuality/index.php?category=' + IntToStr(NewsItem.Category);
      Description := NewsItem.Content + ' (' + NewsItem.Author + ')'; // + EnclosuresText;
      Time := NewsItem.Time;
    end;
    Items.Add(NewItem);
  end;
  end;
  NewsItem.Free;
  News.Free;
  Result := RSSChannel.ProduceOutput;
end;

{ TNews }

constructor TNews.Create;
begin
  NewsCategoryList := TList.Create;
end;

destructor TNews.Destroy;
var
  I: Integer;
begin
  for I := 0 to NewsCategoryList.Count - 1 do
    TNewsCategory(NewsCategoryList[I]).Free;
  NewsCategoryList.Free;
  inherited Destroy;
end;

procedure TNews.LoadFromDb;
var
  DbRows: TDbRows;
  I: Integer;
  NewCategory: TNewsCategory;
begin
  DbRows := Database.Select('news_category', '*');
  for I := 0 to DbRows.Count - 1 do begin
    NewCategory := TNewsCategory.Create;
    NewCategory.LoadFromDbRecord(DbRows[I]);
    NewsCategoryList.Add(NewCategory);
  end;
  DbRows.Free;
end;

{ TNewsCategory }

procedure TNewsCategory.LoadFromDbRecord(DbRow: TAssocArray);
begin
  Id := StrToInt(DbRow.Values['id']);
  Title := DbRow.Values['caption'];
  Permission := StrToInt(DbRow.Values['permission']);
end;

procedure TNewsCategory.Store;
var
  Values: TAssocArray;
begin
  Values := TAssocArray.Create;
  Values.AddKeyValue('permission', IntToStr(Permission));
  Values.AddKeyValue('title', Title);
  Database.Update('news_category', Values, 'id=' + IntToStr(Id));
  Values.Free;
  inherited;
end;

function TNews.ShowNews(Category: Integer): string;
var
  DbRows: TDbRows;
  FontSize: Integer;
  I: Integer;
  II: Integer;
  Enclosures: TArrayOfString;
  Content: string;
const
  NewsCountPerCategory = 5;
//  UploadedFilesFolder = 'aktuality/uploads/';
begin
//  global $NewsCategoryNames, $NewsCountPerCategory, $UploadedFilesFolder;
  Result := '<table class="MainTable"><tr>' +
    '<td>' + TNewsCategory(NewsCategoryList[Category]).Title + '</td><td align="right">'+
    '<a href="?p=news-list&amp;category=' + IntToStr(Category) + '">Zobrazit všechny aktuality</a>&nbsp;';
  if TNewsCategory(NewsCategoryList[Category]).Permission = 1 then
    Result := Result + '<a href="?p=news-add&amp;category=' + IntToStr(Category) + '">Přidat aktualitu</a>&nbsp;';
  Result := Result + '</td></tr><tr><td colspan="2" style="NewsTable">' +
    '<table class="NewsTable" cellspacing="0" cellpadding="0">';
  FontSize := 12;
  DbRows := Database.Select('news', '*', 'category=' + IntToStr(Category) + ' ORDER BY date DESC LIMIT 0,' + IntToStr(NewsCountPerCategory));
  for I := 0 to DbRows.Count - 1 do with DbRows[I] do begin
    Content := Values['content'];
    Content := StringReplace(Content, '<br>', '<br />', [rfReplaceAll]);
    Result := Result + '<tr><td  onclick="window.location=''?p=news-item&amp;id=' + Values['id'] + '''" onmouseover="zobraz(''new' + Values['id'] + ''')" class="NewsTableItemTitle">' +
      '<table class="NewsTableItemTitle">' +
      '<tr><td style="font-size: ' + IntToStr(FontSize) + 'pt"><strong>' + Values['title'] + '</strong></td>'+
      '<td align="right" style="font-size: ' + IntToStr(FontSize) + 'pt">' + Values['author'] + ' (' + HumanDate(Values['date']) + ')' +
      '</td></tr></table>' +
      '<div id="new' + Values['id'] + '" class="NewsTableItem">' + Content;

    if Values['enclosure'] <> '' then begin
      Result := Result + '<br />Přílohy: ';
      Enclosures := Explode(';', Values['enclosure']);
 	    for II := 0 to High(Enclosures) do begin
 	      // if(file_exists($UploadedFilesFolder.$Enclosure)) echo(' <a href="'.$UploadedFilesFolder.$Enclosure.'">'.$Enclosure.'</a>')
      end;
    end;
    Result := Result + '</div></td></tr>';
    FontSize := FontSize - 1;
  end;
  Result := Result + '</table></td></tr></table>';
end;

end.
